from setuptools import setup

setup(
    name='jabutiutils',
    description=u'Jabuti Technologies - Utils',
    version='0.0.4',
    license='MIT License',
    author='Thales Araújo',
    install_requires=['msal>=1.6.0,<2', 'django'],
)